/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: exception.h
 *
 * ABSTRACT: Classes for raising and handling exceptions/failures.
 *
 * EXPORTS:
 *
 * $Revision: 1.7 $
 * $Date: 1999/08/04 14:00:18 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: exception.h,v $
 * Revision 1.7  1999/08/04 14:00:18  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.6  98/09/15  18:45:20  da0g
 * Enhanced exceptions to support multiple-name resolution and Ref_Count (automatically-destroyed) Data.
 * 
 * Revision 1.5  1997/12/29 17:06:17  reids
 * Version that has the basic functionality needed to support TDL.
 *
 * Revision 1.4  97/12/22  16:52:53  reids
 * Basically, added "data" field for CALLBACK_ACTION,
 *  and started using "nodeData" for the activation data associated
 *  with monitors, and the failure data associated with exceptions.
 * 
 * Revision 1.3  97/12/18  00:21:43  reids
 * Changing ACTION_PTR to a handle, to aid in garbage collection.
 * 
 * Revision 1.2  97/12/04  17:50:08  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:30  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 **************************************************************************/

#ifndef INCexception
#define INCexception

#include "taskTree.h"

class Exception_Handler : public Named_Object, public Ref_Count
{
 public:
  Exception_Handler() { action = NULL; maxInvocations = numInvocations = 0; }
  Exception_Handler(STRING type, const TCM_Action_Ref &handler,
		    unsigned int maxTimes)
    : Named_Object(type) { action = handler; maxInvocations = maxTimes;
			   numInvocations = 0; }

  TCM_Action_Ref getAction(void) const { return action; }
  void use (void) { numInvocations++; }
  BOOLEAN matches (STRING failure) const
    { return ((!name || streq(name, failure)) &&
	      (maxInvocations > numInvocations)); }

  BOOLEAN matches ( TCM_StringMatcher theFailureStringMatcher ) const
  {
    return ( ( !name  ||  ( ( * theFailureStringMatcher ) ( name ) == TRUE ) )
	    && ( maxInvocations > numInvocations ) );
  }

 private:
  TCM_Action_Ref action;
  unsigned int maxInvocations;
  unsigned int numInvocations;
};

typedef Exception_Handler *Exception_Handler_Ptr;
typedef Exception_Handler const *Const_Exception_Handler_Ptr;

class Exception_Instance : public Named_Object
{
protected:
  Exception_Handler_Ref     handler;
  Task_Tree_Ref             contextNode;
  TCM_StringMatcher         stringMatcher;

 public:
  Exception_Instance()
    : Named_Object  (                            ),
      handler       (                            ),
      contextNode   (                            ),
      stringMatcher ( TCM_StringMatcher ( NULL ) )
  { }

  Exception_Instance ( const char          * theType,
		       const Task_Tree_Ref & theNode )
    : Named_Object  ( theType                    ),
      handler       (                            ),
      contextNode   ( theNode                    ),
      stringMatcher ( TCM_StringMatcher ( NULL ) )
  { }

  Exception_Instance ( const char        * theId,
		       TCM_StringMatcher   theStringMatcher,
		       Task_Tree_Ref       theNode )
    : Named_Object  ( theId            ),
      handler       (                  ),
      contextNode   ( theNode          ),
      stringMatcher ( theStringMatcher )
  { }

  Exception_Instance ( const Exception_Instance & theExceptionInstance )
    : Named_Object  ( theExceptionInstance               ),
      handler       ( theExceptionInstance.handler       ),
      contextNode   ( theExceptionInstance.contextNode   ),
      stringMatcher ( theExceptionInstance.stringMatcher )
  { }

  ~Exception_Instance() {}

  const Exception_Handler_Ref & getHandler()      const {return handler;      }
  const Task_Tree_Ref         & getContextNode()  const {return contextNode;  }
        TCM_StringMatcher       getStringMatcher()const {return stringMatcher;}

  void  setHandler       (       Exception_Handler * theExceptionHandler )
				   { handler       = theExceptionHandler; }

  void  setContextNode   ( const Task_Tree_Ref     & theContextNode )
				   { contextNode   = theContextNode;      }

  void  setStringMatcher (       TCM_StringMatcher   theStringMatcher )
				   { stringMatcher = theStringMatcher;    }

  void findNext ();

  void raise ( const Task_Tree_Ref & theParent, 
	       const Task_Tree_Ref & theFailureNode,
	       const void          * theFailureData )
	{ contextNode->findAndApplyExceptionHandler(theParent, theFailureNode,
						    this, 
						    theFailureData, FALSE); }

  void raise ( const Task_Tree_Ref & theParent, 
	       const Task_Tree_Ref & theFailureNode,
	       Ref_Count           * theFailureData )
    { contextNode->findAndApplyExceptionHandler(theParent, theFailureNode,
						    this, 
						    theFailureData, TRUE); }

  void addExceptionNode ( const Task_Tree_Ref & theParent, 
			  const Task_Tree_Ref & theFailureNode,
			  const void          * theFailureData,
			  BOOLEAN               theIsRefCountFailureData );
};

class Exception_Node : public Task_Tree_Goal_Node
{
protected:
  Exception_Instance  exceptionInstance;
  Task_Tree_Ref       failureNode;
  BOOLEAN             isRefCountFailureData;

public:
  Exception_Node ( const char * theNodeName )
    : Task_Tree_Goal_Node   ( theNodeName ),
      exceptionInstance     (             ),
      failureNode           (             ),
      isRefCountFailureData ( FALSE       )
  { }

  Exception_Node ( const char               * theNodeName,
		   const TCM_Action_Ref     & theNodeAction,
		   const Exception_Instance & theExceptionInstance,
		   const Task_Tree_Ref      & theFailureNode,
		   BOOLEAN                    theIsRefCountFailureData )
    : Task_Tree_Goal_Node   ( theNodeName, theNodeAction ),
      exceptionInstance     ( theExceptionInstance       ),
      failureNode           ( theFailureNode             ),
      isRefCountFailureData ( theIsRefCountFailureData   )
  { }

  virtual BOOLEAN isException(void) { return TRUE; }

  Task_Tree_Ref getFailureNode() const { return failureNode; }
  Task_Tree_Ref getContextNode() const 
    { return exceptionInstance.getContextNode(); }

  BOOLEAN getIsRefCountFailureData() const  { return isRefCountFailureData; }

  void bypass (void);
};

void raiseException (STRING type, Task_Tree_Ref const &failureNode,
		     const void *failureData);

void raiseException ( TCM_StringMatcher         theType,
		      const char              * theId,
		      const Task_Tree_Ref     & theFailureNode,
		      Ref_Count               * theFailureData );
		     
void addExceptionHandler (Task_Tree_Ref const &nodeRef, STRING type, 
			  const TCM_Action_Ref & action,
			  unsigned int maxInvocations);
void removeExceptionHandler (Task_Tree_Ref const &nodeRef, STRING type,
			     const TCM_Action_Ref & action);

#endif // INCexception
