/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * FILE: external.cc
 *
 * ABSTRACT: Defines functions for dealing with external events.
 *
 * EXPORTS:
 *
 * $Revision: 1.2 $
 * $Date: 1998/10/30 11:16:32 $
 * $Author: da0g $
 *
 * REVISION HISTORY:
 *
 * $Log: external.cc,v $
 * Revision 1.2  1998/10/30 11:16:32  da0g
 * Added ExternalEventBypassHandler.
 *
 * Revision 1.1  1997/12/29 17:06:18  reids
 * Version that has the basic functionality needed to support TDL.
 *
 **************************************************************************/

#ifndef macintosh
#include <errno.h>
#else
#include <sys/errno.h>
#endif
#include "tcmLogging.h"
#include "external.h"
#ifdef macintosh
#include "agenda.h"
#endif


BOOLEAN runDispatchEvents ( void * theExternalEvents,
			 struct timeval * timeout )
{
  return ((External_Events *) theExternalEvents) -> _dispatchEvents( timeout );
}


/**************************************************************************
 *
 * CLASS: External Events
 *
 **************************************************************************/

void External_Events::dispatchEvents (struct timeval *timeout)
{
  if ( getHasBypassHandler() )
    ( * getBypassHandler() ) ( & runDispatchEvents,
			       (void *)this,
			       timeout,
			       getHasDispatchableEvents() );
  else
    _dispatchEvents ( timeout );
}


BOOLEAN External_Events::_dispatchEvents (struct timeval *timeout)
{
  fd_set readMask;
  struct timeval aTimeout, * aTimeoutPtr;
  int    ret;
  
  do {
	/* If select fails, readMask & timeout could be undefined. */
    readMask = _eventMask;

    if ( timeout != (struct timeval *)NULL )
    {
      aTimeout . tv_sec  = timeout -> tv_sec;
      aTimeout . tv_usec = timeout -> tv_usec;
      aTimeoutPtr = & aTimeout;
    }
    else
    {
      aTimeoutPtr = (struct timeval *)NULL;
    }

    ret = select ( _maxConnection+1, &readMask, NULL, NULL, aTimeoutPtr );

  } while (ret < 0 && errno == EINTR);
  
	/* select *might* change the timeout value. */
  if ( timeout != (struct timeval *)NULL )
  {
    timeout -> tv_sec  = aTimeout . tv_sec;
    timeout -> tv_usec = aTimeout . tv_usec;
  }


  if ( ret == 0 )
  { // Timeout occurred
    return FALSE;
  }
  else if ( ret < 0 )
  {
    tcmError("dispatchEvents: Error on select (%d)\n", errno);
    return FALSE;
  }
  else
  {
    for ( int i=0;  i <= _maxConnection; i++ )
    {
      if ( FD_ISSET ( i, & readMask ) )
      {
        _find ( i ) -> invoke();
      }
    }
    return TRUE;
  }
}

void External_Events::addExternalEvent (int sd, SD_CALLBACK_FN_TYPE callback,
					const void *clientData)
{
	External_Event externalEvent(sd, callback, clientData);
	
	if (_remove(sd)) 
	  tcmWarning("addExternalEvent: Replacing existing external event handler for sd %d\n", sd);
	
	_eventList.insertLast(externalEvent);
	FD_SET(sd, &_eventMask);
	if (sd > _maxConnection) _maxConnection = sd;
}
 
void External_Events::removeExternalEvent (int sd)
{	
	if (_remove(sd)) {
		FD_CLR(sd, &_eventMask);
		if (sd == _maxConnection) {
		  // Reset the "maxConnection" value
		  _maxConnection = 0;
			for (int i=sd-1; i>0; i--) {
			  if (FD_ISSET(i, &_eventMask)) {
			  	_maxConnection = i;
			  	break;
			  }
			}
		}
	} else {
	  tcmWarning("removeExternalEvent: No existing external event handler for sd %d\n", sd);
	}
}

External_Event * External_Events::_find (int sd) const
{
	Const_External_Event_Iterate externIter(_eventList);
	External_Event externEvent(sd);
	
	for (; externIter; externIter++) {
		if (*externIter() == externEvent) {
			return externIter();
    }
  }
  return NULL;
}

 
BOOLEAN External_Events::_remove (int sd)
{
	External_Event *externalEvent = _find(sd);
	
	if (externalEvent) {
	  _eventList.removeItem(*externalEvent);
	}
  return externalEvent != NULL;
}
