/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: monitor.cc
 *
 * ABSTRACT: Classes for implementing monitors (point, polling, demon, event)
 *
 * EXPORTS:
 *
 * $Revision: 1.9 $
 * $Date: 1999/08/04 14:00:18 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: monitor.cc,v $
 * Revision 1.9  1999/08/04 14:00:18  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
// Revision 1.8  98/04/21  12:45:58  reids
// Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
//   particular event occurs.
// Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
//   a node by waiting some msecs after a particular event has occurred.
// 
// Revision 1.7  98/01/30  14:50:52  reids
// Updated to compile under gcc 2.7.2 and under Linux.
// Also, made STRING "const char *" and changed API to take const arguments,
//   where applicable.
// 
// Revision 1.6  97/12/30  12:28:12  reids
// Added a "timer" monitor, which is like a polling monitor, except when
//   activated it just invokes a function (rather than adding a new node)
// 
// Revision 1.5  97/12/29  17:06:23  reids
// Version that has the basic functionality needed to support TDL.
// 
// Revision 1.4  97/12/22  16:52:54  reids
// Basically, added "data" field for CALLBACK_ACTION,
//  and started using "nodeData" for the activation data associated
//  with monitors, and the failure data associated with exceptions.
// 
// Revision 1.3  97/12/18  00:21:44  reids
// Changing ACTION_PTR to a handle, to aid in garbage collection.
// 
// Revision 1.2  97/12/04  17:50:12  reids
// Another fairly stable version (except that monitors do not quite work)
// 
// Revision 1.1  97/11/21  14:06:33  reids
// First release of TCM -- seems to be a stable version
// 
 *
 **************************************************************************/

#include <stdio.h>
#include "agenda.h"
#include "tcmGlobal.h"
#include "monitor.h"

/**************************************************************************
 *
 * CLASS: Monitor_Action
 *
 * Description: A specialization of _Action, containing additional information
 *              needed to specify the action of a monitor.
 *
 **************************************************************************/

/**************************************************************************
 *
 * FUNCTION: void activate (Task_Tree_Ref const &node, const void *data)
 *
 * DESCRIPTION: Unless activated enough times already, add a new activation
 *              (command) node as a child of the monitor.
 *
 **************************************************************************/

void Monitor_Action::activate (Task_Tree_Ref const &node)
{
  tcmMessage("Monitor_Action::activate: %d\n", _numActivations);
  if (_numActivations < _maxActivations) {
    _numActivations++;
    Task_Tree_Command_Node *activationNode
      = new Task_Tree_Command_Node(_activationName);
    activationNode->setAction(_activationAction);
    activationNode->setNodeData(node->getNodeData());
    activationNode->addConstraints(_activationConstraints);
    taskTreeInsert(*node, activationNode, TRUE);
   if (_numActivations == _maxActivations) {
      node->doneHandling();
    }
  } else {
    tcmWarning("activate: Trying to activate an inactive monitor\n");
  }
}

/**************************************************************************
 *
 * CLASS: Polling_Monitor_Action
 *
 * Description: A specialization of Monitor_Action, containing additional 
 *              information needed to specify the action of a polling monitor.
 *
 **************************************************************************/

/**************************************************************************
 *
 * FUNCTION: void execute (Task_Tree_Ref const &node)
 *
 * DESCRIPTION: Activate the monitor and set the polling timer.
 *
 **************************************************************************/

void Polling_Monitor_Action::execute (Task_Tree_Ref const &node)
{
  if (!_initialWait) {
    _initialWait = FALSE;
    activate(node);
  }

  if (_numActivations < _maxActivations)
    addTimedEvent(absoluteTime(_period), node, Timeout_Signal);
}

/**************************************************************************
 *
 * CLASS: Task_Tree_Monitor_Node
 *
 * Description: A specialization of Task_Tree_Command_Node, containing 
 *              additional information needed to specify a monitor.
 *
 **************************************************************************/

/**************************************************************************
 *
 * FUNCTION: void trigger (void)
 *
 * DESCRIPTION: Increment the number of times this monitor has been 
 *              triggered.  Complete handling the monitor if it has been
 *              triggered enough times, and is not already done (e.g.,
 *              may have been activated maxActivation times already).
 *
 **************************************************************************/

void Task_Tree_Monitor_Node::trigger (void)
{
  _numTriggers++;
  if (_numTriggers == _maxTriggers) {
    if (!validState(Handled_State)) doneHandling();
  }
}

/**************************************************************************
 *
 * FUNCTION: void activate (const void *data)
 *
 * DESCRIPTION: Activate the monitor (invoking the activation action).
 *
 **************************************************************************/

void Task_Tree_Monitor_Node::activate (const void *data)
{
  if (isCurrentState(Handling_State)) {
    setNodeData(data);
    ((Monitor_Action *)(*_action))->activate(this);
  } else {
    tcmWarning("Trying to activate an inactive monitor %s\n", name);
  }
}

/**************************************************************************
 *
 * FUNCTION: void _handleSignal(Signal_Enum signal)
 *
 * DESCRIPTION: A signal just arrived at the monitor; Handle it.
 *              Differs from handling signals for regular nodes in
 *              that it handles Activate and Timeout signals.
 *
 **************************************************************************/

void Task_Tree_Monitor_Node::_handleSignal(Signal_Enum signal)
{
  switch (signal) {
  case Activate_Signal: 
    tcmMessage("Task_Tree_Monitor_Node::_handleSignal: Activation\n");
    activate(_nodeData);
    break;

  case Timeout_Signal: 
    tcmMessage("Task_Tree_Monitor_Node::_handleSignal: Timeout\n");
    if (isCurrentState(Handling_State)) {
      execute();
    } else {
      tcmWarning("Trying to execute an inactive monitor %s\n", name);
    }
    break;
    
  case Start_Terminating_Signal:
    // Force node to be handled, so that it terminates correctly
    if (!validState(Handled_State)) doneHandling();
    Task_Tree_Command_Node::_handleSignal(signal);
    break;
    
  case Done_Handling_Signal:
		Task_Tree_Command_Node::_handleSignal(signal);
		// Remove all activation and timeout signals from handled monitors
		_finishHandlingMonitor();		
		 break;

  default:
    Task_Tree_Command_Node::_handleSignal(signal); break;
  }
}

/**************************************************************************
 *
 * FUNCTION: void _finishHandlingMonitor(void)
 *
 * DESCRIPTION: Remove all activation and timeout signals from handled monitors
 *
 **************************************************************************/

void Task_Tree_Monitor_Node::_finishHandlingMonitor (void)
{
  Event_Iterator expectIter(expectedEvents);
  Task_Tree_Node_Ptr expectNode = NULL;
  Signal_Enum expectSignal;
	
  // I don't think the current list implementation will work right 
  // if "removeRequestedEvents" removes something from the list being
  // iterated on here.  This is wasteful, but should work all the time.
  for (expectIter.reset(); expectIter; (expectNode == this 
					? (expectNode=NULL, expectIter.reset())
					: expectIter++)) {
    expectSignal = expectIter()->getSignal();
    if (expectSignal == Activate_Signal || expectSignal == Timeout_Signal) {
      expectNode = *(expectIter()->getNode());
      removeRequestedEvents(*expectIter());
      expectIter.removeCurrent();
    }
  }
}

/**************************************************************************
 *
 * Public Functions
 *
 **************************************************************************/

/**************************************************************************
 *
 * FUNCTION: Task_Tree_Monitor_Node *monitorDereference (TCM_Task_Tree_Ref const &nodeRef)
 *
 * DESCRIPTION: Return the monitor node (either the node itself, or its parent)
 *
 **************************************************************************/

Task_Tree_Monitor_Node *monitorDereference (TCM_Task_Tree_Ref const &nodeRef)
{
  if (nodeRef->isMonitor()) {
    return (Task_Tree_Monitor_Node *)*nodeRef;
  } else if (nodeRef->getParent()->isMonitor()) {
    return (Task_Tree_Monitor_Node *)*nodeRef->getParent();
  } else {
    return NULL;
  }
}
