/**************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: taskTree.h
 *
 * ABSTRACT: Defines the hierarchy of task tree nodes and operations
 *           to create them, link them into a tree, and execute actions.
 *
 * EXPORTS:
 *
 * $Revision: 1.13 $
 * $Date: 2000/01/19 21:26:42 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: taskTree.h,v $
 * Revision 1.13  2000/01/19 21:26:42  reids
 * Added two new top-level functions:
 *   TCM_IsDoneHandling(ref) -- returns TRUE if the ref has raised success or
 * 			     failure.
 *   TCM_IsPostponed(ref) -- returns TRUE if the ref's action has finished
 * 		          executing, but the node has not yet raised
 * 			  success or failure.
 *
 * Revision 1.12  1999/08/04 14:00:19  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.11  1999/06/06 13:48:09  reids
 * Changes to TCM 1.3.0:
 *   New list data structures.
 *   New, generalized logging interface.
 *   Fixed firstChild to work if there are no children.
 *   Fixed code to deallocate a node.
 *   In certain circmumstances, don't send termination signal when task
 *     itself terminating.
 *
 * Revision 1.10  1998/12/16 03:08:19  reids
 * Added support for "on termination" functions.
 *   Also enabled tca.h and tcm.h to co-exist (needed to change values
 *   of several constants).
 *
 * Revision 1.9  98/09/15  18:45:22  da0g
 * Enhanced exceptions to support multiple-name resolution and Ref_Count (automatically-destroyed) Data.
 * 
 * Revision 1.8  1998/08/05 10:57:26  reids
 * Made a compiler option for printing out debugging information.
 *   Created macros for TCM_TerminateAtAfter and TCM_DelayUntilAfter, for
 *   backwards compatibility.
 *
 * Revision 1.7  98/04/21  12:46:07  reids
 * Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
 *   particular event occurs.
 * Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
 *   a node by waiting some msecs after a particular event has occurred.
 * 
 * Revision 1.6  97/12/30  12:26:50  reids
 * Added flag to indicate whether node raised an exception (ie, failed).
 * 
 * Revision 1.5  97/12/29  17:06:29  reids
 * Version that has the basic functionality needed to support TDL.
 * 
 * Revision 1.4  97/12/22  16:53:01  reids
 * Basically, added "data" field for CALLBACK_ACTION,
 *  and started using "nodeData" for the activation data associated
 *  with monitors, and the failure data associated with exceptions.
 * 
 * Revision 1.3  97/12/18  00:21:47  reids
 * Changing ACTION_PTR to a handle, to aid in garbage collection.
 * 
 * Revision 1.2  97/12/04  17:50:19  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:38  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 **************************************************************************/

#ifndef INCtaskTree
#define INCtaskTree

#include <stdio.h>
#include "tcmLogging.h"
#include "event.h"
#include "action.h"

//#define TRACE
#ifdef TRACE
//#define TRACE_SIGNALS
#define TRACE_TRANSITIONS
#define TRACE_DELETIONS
#define TRACE_EXCEPTIONS
#define TRACE_TIMER
#define TRACE_ACTION_DELETIONS
//#define TRACE_VIRTUAL
#define TRACE_AFTER_SIGNAL
#endif

#ifdef DISTRIBUTED
typedef struct {
  STRING       host;
  STRING       creator;
  unsigned int id;
} Virtual_Address;

#define DVIRTUAL virtual
#else
#define DVIRTUAL
#endif

inline BOOLEAN immediateSignal (Signal_Enum signal) 
{ return signal == Activate_Signal || signal == Start_Terminating_Signal; }

typedef list<Task_Tree_Ref> Task_Tree_Node_List;
typedef List_Iterate<Task_Tree_Ref> Task_Tree_Node_Iterator;
typedef Const_List_Iterate<Task_Tree_Ref> Const_Task_Tree_Node_Iterator;

class Exception_Handler;  // Forward reference, to keep compiler happy
class Exception_Instance; // Forward reference, to keep compiler happy

typedef tcmHandle<Exception_Handler> Exception_Handler_Ref;

typedef list<Exception_Handler_Ref> Exception_List;
typedef List_Iterate<Exception_Handler_Ref> Exception_Iterator;
typedef Const_List_Iterate<Exception_Handler_Ref> Const_Exception_Iterator;
typedef Exception_List *Exception_List_Ptr;
typedef Exception_List const *Const_Exception_List_Ptr;

typedef Task_Tree_Node *Task_Tree_Node_Ptr;

typedef tcmHandle<Ref_Count>  TCM_Generic_Ref;

class Task_Tree_Node : public Ref_Count, public Named_Object
{
 public:
  Task_Tree_Node() : Named_Object("Unknown")
    { _initialize(); _intraNodeConstraints(); }
  Task_Tree_Node(STRING nodeName) : Named_Object(nodeName)
    { _initialize(); _intraNodeConstraints(); }
  Task_Tree_Node(STRING nodeName, const TCM_Action_Ref &nodeAction)
    : Named_Object(nodeName)
    { _initialize(); _action = nodeAction; _intraNodeConstraints(); }
  ~Task_Tree_Node() {
#ifdef TRACE_DELETIONS
    tcmMessage("Deleting %s\n", name);
#endif
#ifdef DISTRIBUTED
    if (validVirtualAddress()) {
      delete[] _vaddress.host;  delete[] _vaddress.creator;
    }
#endif
  }

  Task_Tree_Ref getParent(void) const { return parent; }
  Exception_List_Ptr getExceptions(void) { return &exceptions; }
  Const_Exception_List_Ptr getExceptions(void) const { return &exceptions; }

  virtual BOOLEAN isGoal(void)      { return FALSE; }
  virtual BOOLEAN isException(void) { return FALSE; }
  virtual BOOLEAN isMonitor(void)   { return FALSE; }

  DVIRTUAL BOOLEAN    validState(State_Enum state) const ;
  DVIRTUAL BOOLEAN    isCurrentState(State_Enum state) const ;
  DVIRTUAL State_Enum theCurrentState(State_Enum state) const
    { return currentState.current(state); }
  DVIRTUAL BOOLEAN    isChild(Task_Tree_Ref const &childNode) const 
    { return children.member(childNode); }

  DVIRTUAL void addParent(Task_Tree_Ref const &parentNode,
			  BOOLEAN canBeLastChild,
			  BOOLEAN addConstraints=TRUE);
  DVIRTUAL void addChild(Task_Tree_Ref const &childNode,
			 BOOLEAN addConstraints=TRUE);
  DVIRTUAL void removeChild(Task_Tree_Ref const &childRef)
    { children.removeItem(childRef); }

  DVIRTUAL Task_Tree_Node_Ptr firstChild(void) const;
  DVIRTUAL Task_Tree_Node_Ptr lastChild(BOOLEAN ignoreVeryLast) const;
  DVIRTUAL Task_Tree_Node_Ptr previousChild(Task_Tree_Ref const &child) const;
  DVIRTUAL Task_Tree_Node_Ptr nextChild(Task_Tree_Ref const &child) const;
  DVIRTUAL Task_Tree_Node_Ptr childNamed(STRING const name) const;
  DVIRTUAL Task_Tree_Node_Ptr ancestorNamed(STRING const name) const;

  DVIRTUAL void execute (void) {
    if (*_action) {
      _action->execute(this);
      if (!_doneHandling) _postponed = TRUE;
    } else tcmError("No action for node %s\n", this->name);
  }
  DVIRTUAL void signal(const Event &event);

  DVIRTUAL void addExpectedEvent(Signal_Enum signal,
				 Task_Tree_Ref const &signallingNode,
				 const void *timerData = NULL);
  DVIRTUAL void addRequestedEvent(State_Enum state, Signal_Enum signal,
				  Task_Tree_Ref const &signalledNode);
  DVIRTUAL void addRequestedEvent(State_Enum state,
				  SIMPLE_CALLBACK_FN_TYPE callbackFn,
				  const void *data);

  DVIRTUAL void removeExpectedEvent(Signal_Enum signal,
				    Task_Tree_Ref const &signallingNode);
  DVIRTUAL void removeRequestedEvents(Signal_Enum signal,
				      Task_Tree_Ref const &signalledNode);
  void removeRequestedEvents(const Event &event);
  DVIRTUAL void processRequestedEvents(void);
  DVIRTUAL void transitionTo(State_Enum newState);
  DVIRTUAL void doneHandling(void);

  DVIRTUAL void setFailed (void) { _failed = TRUE; }

  // Return TRUE if failed, FALSE otherwise
  DVIRTUAL BOOLEAN getNodeFailed (void) const { return _failed; }
  // Return TRUE if any node of the tree failed, FALSE otherwise
  DVIRTUAL BOOLEAN getTreeFailed (void) const;

  DVIRTUAL BOOLEAN terminate (void);
  DVIRTUAL BOOLEAN deallocate (void);
  DVIRTUAL void suspend (void);
  DVIRTUAL void unsuspend (void);
  DVIRTUAL BOOLEAN isSuspended (void) const { return _suspended; }

  DVIRTUAL BOOLEAN isDoneHandling (void) const { return _doneHandling; }
  DVIRTUAL BOOLEAN isPostponed    (void) const { return _postponed; }

  DVIRTUAL void delayTermination  (void) { _delayedTermination = TRUE; }
  DVIRTUAL void enableTermination (void) { _delayedTermination = FALSE; }
  DVIRTUAL BOOLEAN isTerminationDelayed (void) const
    { return _delayedTermination; }
  
  // Defined in tplConstr.cc
  void addConstraints(int tplConstraints);

  virtual STRING className(void) const { return "Node"; }
  virtual STRING instanceName(void) const { return getName(); }

  int getNodeId (void) const { return id; }

  DVIRTUAL void setNodeData (void const *nodeData) { _nodeData = nodeData; }
  DVIRTUAL void const *getNodeData (void) const { return _nodeData; }

  DVIRTUAL Ref_Count * getNodeRefCountData() const
    { return _nodeRefCountData . operator*();  }
  DVIRTUAL void setNodeRefCountData ( Ref_Count * theNodeRefCountData )
    { _nodeRefCountData = theNodeRefCountData; }

  DVIRTUAL void setAction (TCM_Action_Ref action) { _action = action; }
  DVIRTUAL TCM_Action_Ref getAction (void) const { return _action; }

  // Functions to map between a task tree node and its id number
  int remember (void);
  Task_Tree_Node_Ptr recall (int refID) const;
  void forget (int refID);
  Task_Tree_Node_Ptr retrieve (int refID)
    { Task_Tree_Node_Ptr node = recall(refID); forget(refID); return node; }

  DVIRTUAL void displayTree (FILE *stream, int position = 0) const;

  DVIRTUAL void addTerminationAction (const TCM_Task_Tree_Ref &terminationRef);

  DVIRTUAL void findAndApplyExceptionHandler (Task_Tree_Ref parentNode,
					      Task_Tree_Ref failureNode,
					      Exception_Instance *exception,
					      const void *data,
					      BOOLEAN isRefCountFailureData);
  DVIRTUAL void noExceptionHandler (STRING failure) const;

  DVIRTUAL void addTimedEventAfterSignal (MSecs waitTime,
					  const Task_Tree_Ref nodeRef, 
					  Signal_Enum signal,
					  State_Enum state);

 private: // Data items
  static int globalId;
  int id;
  State currentState;
  Task_Tree_Ref parent;
  Task_Tree_Node_List children;
  BOOLEAN canBeLast;
  BOOLEAN _suspended;
  BOOLEAN _failed; // TRUE if this node raised an exception
  // TRUE if this node is done handling (raised success or failure)
  BOOLEAN _doneHandling;
  // TRUE if handler code has run, but task has not declared success or failure
  BOOLEAN _postponed;
  // TRUE if node is an "on termination node (or one of its descendants)
  BOOLEAN _delayedTermination;

  Exception_List exceptions;
  
  Task_Tree_Node_List terminationRefs;

  static Task_Tree_Node_List memory;

 protected: // Data items
  Event_List expectedEvents;
  State_Event_List requestedEvents;
  
  TCM_Action_Ref    _action;
  void const      * _nodeData;
  TCM_Generic_Ref   _nodeRefCountData;

 protected: // Functions
  DVIRTUAL /*inline*/ BOOLEAN _noneExpected(Signal_Enum signal) const;
  // Different classes of nodes may handle the same signal in different ways.
  virtual void _handleSignal (Signal_Enum signal);
  virtual void _intraNodeConstraints (void);

 private: // Functions
   // Initialize task tree node data structure
  DVIRTUAL void _initialize(void);
  DVIRTUAL void _processSignal (Signal_Enum signal) {
    // Handle the signal if no other signals of the given type are expected,
    //  or if the signal is to terminate the node (since this should happen
    //  if *any* event wants to terminate the node
    if (immediateSignal(signal) || _noneExpected(signal))
      _handleSignal(signal);
  }

  DVIRTUAL BOOLEAN _handleTransition (State_Enum newState);
  DVIRTUAL void _startTerminating (void);
  DVIRTUAL void _destroy (void);

  //  Unsuspend a single node, and trigger any waiting transitions
  DVIRTUAL void _unsuspendNode (void);
  
  DVIRTUAL STRING _nodeStatusString (void) const;

  DVIRTUAL void _forceTimeouts (void);

#ifdef DISTRIBUTED
 protected:
  Virtual_Address _vaddress;
  static Task_Tree_Node_List virtualNodes;
 public:
  BOOLEAN validVirtualAddress (void) { return _vaddress.host != NULL; }
  Virtual_Address &getVirtualAddress (void) { return _vaddress; }
  void setVirtualAddress (STRING host, STRING creator, unsigned int id);
  void setVirtualAddress (const Virtual_Address &vaddress) 
    { setVirtualAddress(vaddress.host, vaddress.creator, vaddress.id); }

  // Functions to map between a task tree node and its virtual address
  void                      rememberVirtual (void);
  void                      forgetVirtual   (const Virtual_Address &vaddress);
  static Task_Tree_Node_Ptr recallVirtual   (const Virtual_Address &vaddress);
  static Task_Tree_Node_Ptr getVirtual      (const Virtual_Address &vaddress,
					     STRING nodeName = "Unknown");
#endif
};

class Task_Tree_Goal_Node : public Task_Tree_Node
{
 public:
  Task_Tree_Goal_Node(STRING nodeName);
  Task_Tree_Goal_Node(STRING nodeName, const TCM_Action_Ref &nodeAction);

  virtual STRING className(void) const { return "Goal"; }
  virtual BOOLEAN isGoal(void) { return TRUE; }

 protected:
  virtual void _handleSignal (Signal_Enum signal);
  virtual void _intraNodeConstraints (void);
};

class Task_Tree_Command_Node : public Task_Tree_Node
{
 public:
  Task_Tree_Command_Node(STRING nodeName);
  Task_Tree_Command_Node(STRING nodeName, const TCM_Action_Ref &nodeAction);
  
  virtual STRING className(void) const { return "Command"; }

 protected:
  virtual void _handleSignal (Signal_Enum signal);
  virtual void _intraNodeConstraints (void);
};

void addEventPair (Task_Tree_Ref const &signallingNode, State_Enum state,
		   Signal_Enum signal, Task_Tree_Ref const &signalledNode);

void taskTreeInsert (Task_Tree_Node_Ptr parent, Task_Tree_Node_Ptr child,
		     BOOLEAN isLastChild);

#endif /* INCtaskTree */
