/*****************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: tcm.h
 *
 * ABSTRACT: Public include file for the Task Control Management library.
 *           Defines the TCM API.
 *
 * $Source: /afs/cs/project/TCA/Master/tcaV8/tcm/src/tcm.h,v $ 
 * $Revision: 1.19 $
 * $Date: 2000/02/02 22:09:47 $
 * $Author: da0g $
 *
 * REVISION HISTORY:
 *
 * $Log: tcm.h,v $
 * Revision 1.19  2000/02/02 22:09:47  da0g
 * Fixed improper templating bugs with tcmHandle's refIncr and refDecr
 *
 * Revision 1.18  2000/01/19 21:26:43  reids
 * Added two new top-level functions:
 *   TCM_IsDoneHandling(ref) -- returns TRUE if the ref has raised success or
 * 			     failure.
 *   TCM_IsPostponed(ref) -- returns TRUE if the ref's action has finished
 * 		          executing, but the node has not yet raised
 * 			  success or failure.
 *
 * Revision 1.17  1999/08/04 14:00:19  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.16  1998/12/21 17:56:13  reids
 * Added TCM_GetAction.
 *
 * Revision 1.15  1998/12/16 03:08:20  reids
 * Added support for "on termination" functions.
 *   Also enabled tca.h and tcm.h to co-exist (needed to change values
 *   of several constants).
 *
 * Revision 1.14  98/12/14  15:37:06  da0g
 * tcm.h
 * 
 * Revision 1.13  98/12/14  15:20:08  da0g
 * Added relativeTimeout to TCM_ProcessAgenda
 * 
 * Revision 1.12  98/10/30  11:16:33  da0g
 * Added ExternalEventBypassHandler.
 * 
 * Revision 1.11  1998/09/15 18:45:23  da0g
 * Enhanced exceptions to support multiple-name resolution and Ref_Count (automatically-destroyed) Data.
 *
 * Revision 1.10  1998/08/05 10:57:28  reids
 * Made a compiler option for printing out debugging information.
 *   Created macros for TCM_TerminateAtAfter and TCM_DelayUntilAfter, for
 *   backwards compatibility.
 *
 * Revision 1.9  98/07/14  17:21:04  reids
 * Changed TCM_DelayUntilAfter to TCM_DelayForAfter, changed TCM_
 *   TerminateAtAfter to TCM_TerminateInAfter to TCM_DelayForAfter, and
 *   changed order of arguments.
 * Added TCM_ActivateInAfter.
 * 
 * Revision 1.8  98/04/21  12:46:13  reids
 * Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
 *   particular event occurs.
 * Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
 *   a node by waiting some msecs after a particular event has occurred.
 * 
 * Revision 1.7  98/01/30  14:51:04  reids
 * Updated to compile under gcc 2.7.2 and under Linux.
 * Also, made STRING "const char *" and changed API to take const arguments,
 *   where applicable.
 * 
 * Revision 1.6  97/12/30  12:29:05  reids
 * Added option to *not* wait for timeouts and external events.
 *   Added flag to indicate whether node raised an exception (ie, failed).
 *   Added a "timer" monitor, which is like a polling monitor, except when
 *     activated it just invokes a function (rather than adding a new node)
 * 
 * Revision 1.5  97/12/29  17:06:33  reids
 * Version that has the basic functionality needed to support TDL.
 * 
 * Revision 1.4  97/12/22  16:53:04  reids
 * Basically, added "data" field for CALLBACK_ACTION,
 *  and started using "nodeData" for the activation data associated
 *  with monitors, and the failure data associated with exceptions.
 * 
 * Revision 1.3  97/12/18  00:21:50  reids
 * Changing ACTION_PTR to a handle, to aid in garbage collection.
 * 
 * Revision 1.2  97/12/04  17:50:22  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:40  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 *****************************************************************************/

#ifndef INCtcm
#define INCtcm

#include <stdio.h>
#include "tcmBasics.h"
#include "tcmHandle.h"

DECLARE_REF_COUNT_FUNCTIONS(_Action);
DECLARE_REF_COUNT_FUNCTIONS(Exception_Handler);
DECLARE_REF_COUNT_FUNCTIONS(Task_Tree_Node);
DECLARE_REF_COUNT_FUNCTIONS(Ref_Count);

enum TCM_Return_Type { TCM_Ok, TCM_Error };

/*****************************************************************
 * 
 * Types and functions having to do with creating task tree nodes
 *   and defining handlers/callbacks.
 *
 *****************************************************************/

typedef tcmHandle<class Task_Tree_Node> TCM_Task_Tree_Ref;

typedef void (*SIMPLE_CALLBACK_FN_TYPE)(const void *callbackData);

typedef void (*CALLBACK_FN_TYPE)(const TCM_Task_Tree_Ref &nodeRef,
				 const void *callbackData);

typedef void (*EXT_CALLBACK_FN_TYPE)(const TCM_Task_Tree_Ref &nodeRef,
				     const void *callbackData,
				     const void *classData);

class _Action : public Ref_Count
{
 public:
  virtual ~_Action(); 

#ifndef macintosh
  virtual void execute (const TCM_Task_Tree_Ref &node) = 0;
#else
  virtual void execute (const TCM_Task_Tree_Ref &node);
#endif
  virtual BOOLEAN operator==(const _Action &otherAction) const 
    { return this == &otherAction; }
};

typedef tcmHandle<_Action> TCM_Action_Ref;

TCM_Task_Tree_Ref const &TCM_RootNode (void);

TCM_Task_Tree_Ref TCM_AllocateGoalNode (STRING name);

TCM_Task_Tree_Ref TCM_AllocateCommandNode (STRING name);

TCM_Return_Type TCM_DeallocateNode (const TCM_Task_Tree_Ref &nodeRef);

TCM_Return_Type TCM_SetAction (const TCM_Task_Tree_Ref &nodeRef, 
			       const TCM_Action_Ref action);

TCM_Action_Ref TCM_GetAction (const TCM_Task_Tree_Ref &nodeRef);

TCM_Return_Type TCM_SetData (const TCM_Task_Tree_Ref &nodeRef,
			     const void *data);

TCM_Return_Type TCM_InsertNode (const TCM_Task_Tree_Ref &parentRef,
				const TCM_Task_Tree_Ref &childRef,
				BOOLEAN isLastChild = TRUE);

// These are shorthands for a series of allocate/setAction/insert calls

TCM_Task_Tree_Ref TCM_CreateGoalNode (const TCM_Task_Tree_Ref &parentRef,
				      STRING name,
				      const TCM_Action_Ref &action,
				      BOOLEAN isLastChild = TRUE);

TCM_Task_Tree_Ref TCM_CreateGoalNode (const TCM_Task_Tree_Ref &parentRef,
				      STRING name, 
				      const void *callbackData,
				      CALLBACK_FN_TYPE callback,
				      BOOLEAN isLastChild = TRUE);

TCM_Task_Tree_Ref TCM_CreateCommandNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name,
					 const TCM_Action_Ref &action,
					 BOOLEAN isLastChild = TRUE);

TCM_Task_Tree_Ref TCM_CreateCommandNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name, 
					 const void *callbackData,
					 CALLBACK_FN_TYPE callback,
					 BOOLEAN isLastChild = TRUE);

TCM_Return_Type TCM_Success (const TCM_Task_Tree_Ref &node);

TCM_Return_Type TCM_Failure (const TCM_Task_Tree_Ref &node, 
			     STRING type, void *failureData = NULL);

TCM_Return_Type TCM_Failure (const TCM_Task_Tree_Ref & theNodeRef,
			     TCM_StringMatcher         theStringMatcher,
			     const char              * theFailureId,
			     Ref_Count               * theFailureData );

/**************************************************************************
 *
 * Functions to map between a task tree node and its id number
 *
 **************************************************************************/

int TCM_Remember (const TCM_Task_Tree_Ref &nodeRef);

TCM_Task_Tree_Ref TCM_Recall (int refID);

void TCM_Forget (int refID);

// Combines TCM_Recall & TCM_Forget
TCM_Task_Tree_Ref TCM_Retrieve (int refID);

/*****************************************************************
 * 
 * Types and functions having to do with temporal constraints
 *
 *****************************************************************/

enum TCM_Interval_Enum { Unknown_Interval, Handling_Interval, 
			   Planning_Interval, Achieving_Interval};

enum TCM_Point_Enum { Unknown_Point, Start_Point, End_Point};

#ifndef INCtca // Need this to enable tca.h and tcm.h to both be included
const int SEQ_ACH        = 1;
const int SEQ_PLANNING   = 2;
const int PLAN_FIRST     = 4;
const int DELAY_PLANNING = 8;
#endif

class TCM_Interval
{
  friend class TCM_Point;
 public:
  TCM_Interval(const TCM_Interval_Enum interval, const TCM_Task_Tree_Ref &node)
    : _which(interval), _ref(node) { };

 protected:
  TCM_Interval_Enum _which;
  TCM_Task_Tree_Ref _ref;
};

class TCM_Point
{
 public:
  TCM_Point(const TCM_Point_Enum point, const TCM_Interval &interval)
    : _which(point), _interval(interval) {};
  BOOLEAN valid (void) const;
  TCM_Point_Enum point(void) const { return _which; }
  TCM_Interval_Enum interval(void) const { return _interval._which; }
  TCM_Task_Tree_Ref node(void) const { return _interval._ref; }

 protected:
  TCM_Point_Enum _which;
  TCM_Interval _interval;
};

TCM_Interval TCM_HandlingOf (const TCM_Task_Tree_Ref &nodeRef);
TCM_Interval TCM_PlanningOf (const TCM_Task_Tree_Ref &nodeRef);
TCM_Interval TCM_AchievingOf (const TCM_Task_Tree_Ref &nodeRef);

TCM_Point TCM_StartOf (const TCM_Interval &interval);
TCM_Point TCM_EndOf (const TCM_Interval &interval);

TCM_Return_Type TCM_AddConstraint (const TCM_Task_Tree_Ref &nodeRef,
				   int tplConstraints);

// Planning and Achievement of "beforeRef" end before start of "afterRef"
TCM_Return_Type TCM_Serialize (const TCM_Task_Tree_Ref &beforeRef,
			       const TCM_Task_Tree_Ref &afterRef);

TCM_Return_Type TCM_DelayUntil (TCM_Point const &timePoint,
				TCM_Point const &delayTimePoint);
TCM_Return_Type TCM_DelayUntil (TCM_Point const &timePoint,
				MSecs absoluteTime);
TCM_Return_Type TCM_DelayFor (TCM_Point const &timePoint,
			      MSecs relativeWait);
// Delay this timepoint for "relativeWait" msecs after 
//  the "delayTimePoint" have passed.
TCM_Return_Type TCM_DelayForAfter (TCM_Point const &timePoint,
				   MSecs relativeWait,
				   TCM_Point const &delayTimePoint);
// For backwards compatibility
#define TCM_DelayUntilAfter(tp, dtp, rw) TCM_DelayForAfter(tp, rw, dtp)

// Don't enable this timepoint until TCM_Signal is invoked for that timepoint
TCM_Return_Type TCM_DelayUntilSignal (TCM_Point const &timePoint);

// Send an explicit signal to enable this timepoint
TCM_Return_Type TCM_Signal (TCM_Point const &timePoint);

TCM_Return_Type TCM_WaitUntil (TCM_Point const &waitTimePoint);
TCM_Return_Type TCM_WaitUntil (MSecs absoluteTime);
TCM_Return_Type TCM_WaitFor   (MSecs relativeWait);

BOOLEAN TCM_IsPast (TCM_Point const &timePoint);
BOOLEAN TCM_IsPast (MSecs absoluteTime);

TCM_Task_Tree_Ref TCM_CreateDelayCommand (const TCM_Task_Tree_Ref &parentRef,
					  MSecs relativeWait);

/*****************************************************************
 * 
 * Functions having to do with traversing task trees.
 *
 *****************************************************************/

TCM_Task_Tree_Ref TCM_Parent (const TCM_Task_Tree_Ref &childRef);
TCM_Task_Tree_Ref TCM_FirstChild (const TCM_Task_Tree_Ref &parentRef);
TCM_Task_Tree_Ref TCM_LastChild (const TCM_Task_Tree_Ref &parentRef);
TCM_Task_Tree_Ref TCM_NextChild (const TCM_Task_Tree_Ref &childRef);
TCM_Task_Tree_Ref TCM_PreviousChild (const TCM_Task_Tree_Ref &childRef);

TCM_Task_Tree_Ref TCM_ChildNamed (const TCM_Task_Tree_Ref &parentRef,
				  STRING childName);
TCM_Task_Tree_Ref TCM_AncestorNamed (const TCM_Task_Tree_Ref &descendantRef,
				     STRING ancestorName);

STRING TCM_NodeName (const TCM_Task_Tree_Ref &nodeRef);

/*****************************************************************
 * 
 * Functions having to do with terminating and suspending nodes
 *
 *****************************************************************/

TCM_Return_Type TCM_TerminateNode (const TCM_Task_Tree_Ref &nodeRef);
TCM_Return_Type TCM_SuspendNode (const TCM_Task_Tree_Ref &nodeRef);
TCM_Return_Type TCM_UnsuspendNode (const TCM_Task_Tree_Ref &nodeRef);
BOOLEAN TCM_IsSuspended (const TCM_Task_Tree_Ref &nodeRef);

BOOLEAN TCM_IsDoneHandling (const TCM_Task_Tree_Ref &nodeRef);
BOOLEAN TCM_IsPostponed    (const TCM_Task_Tree_Ref &nodeRef);

TCM_Return_Type TCM_TerminateAt (const TCM_Task_Tree_Ref &nodeRef,
				 TCM_Point const &terminateTimePoint);
TCM_Return_Type TCM_TerminateAt (const TCM_Task_Tree_Ref &nodeRef,
				 MSecs absoluteTime);
TCM_Return_Type TCM_TerminateIn (const TCM_Task_Tree_Ref &nodeRef,
				 MSecs relativeWait);
// Terminate this node in "relativeWait" msecs after 
//  the "terminateTimePoint" have passed.
TCM_Return_Type TCM_TerminateInAfter (const TCM_Task_Tree_Ref &nodeRef,
				      MSecs relativeWait,
				      TCM_Point const &terminateTimePoint);
// For backwards compatibility
#define TCM_TerminateAtAfter(nr, ttp, rw) TCM_TerminateInAfter(nr, rw, ttp)

TCM_Return_Type TCM_OnTermination (const TCM_Task_Tree_Ref &nodeRef,
				   const TCM_Task_Tree_Ref &terminationRef);

TCM_Return_Type TCM_OnTermination (const TCM_Task_Tree_Ref &nodeRef,
				   STRING name,
				   const TCM_Action_Ref &terminationAction);

TCM_Return_Type TCM_OnTermination (const TCM_Task_Tree_Ref &nodeRef,
				   STRING name,
				   const void *callbackData,
				   CALLBACK_FN_TYPE callback);

/*****************************************************************
 * 
 * Functions having to do with exceptions
 *
 *****************************************************************/

TCM_Return_Type TCM_AddExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					STRING type,
					const TCM_Action_Ref &action,
					unsigned int maxInvocations = MAXINT);

TCM_Return_Type TCM_AddExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					STRING type,
					const CALLBACK_FN_TYPE callback,
					unsigned int maxInvocations = MAXINT);

TCM_Return_Type TCM_RemoveExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					   STRING type,
					   const TCM_Action_Ref &action);

TCM_Return_Type TCM_RemoveExceptionHandler(const TCM_Task_Tree_Ref &nodeRef,
					   STRING type,
					   const CALLBACK_FN_TYPE callback);

TCM_Return_Type TCM_Bypass(const TCM_Task_Tree_Ref &nodeRef);

TCM_Return_Type TCM_Retry(const TCM_Task_Tree_Ref &nodeRef,
			  const void *retryData);

const void *TCM_FailureData (const TCM_Task_Tree_Ref &nodeRef);

/*****************************************************************
 * 
 * Functions having to do with monitors
 *
 *****************************************************************/

TCM_Task_Tree_Ref TCM_AllocateMonitorNode (STRING name);

TCM_Return_Type 
TCM_SetMonitorParameters (const TCM_Task_Tree_Ref &monitorRef,
			  const TCM_Action_Ref &activationAction,
			  unsigned int maxActivations = MAXINT,
			  unsigned int maxTriggers = MAXINT,
			  STRING activationName = NULL,
			  int activationConstraints = 0);

TCM_Task_Tree_Ref TCM_AllocatePollingMonitorNode (STRING name);

TCM_Return_Type 
TCM_SetPollingMonitorParameters (const TCM_Task_Tree_Ref &monitorRef,
				 const TCM_Action_Ref &activationAction,
				 MSecs period = seconds(1),
				 unsigned int maxActivations = MAXINT,
				 unsigned int maxTriggers = MAXINT,
				 STRING activationName = NULL,
				 int activationConstraints = 0,
				 BOOLEAN initialWait = FALSE);

TCM_Task_Tree_Ref TCM_CreateMonitorNode(const TCM_Task_Tree_Ref &parentRef,
					STRING name, 
					const TCM_Action_Ref &activationAction,
					unsigned int maxActivations = MAXINT,
					unsigned int maxTriggers = MAXINT,
					STRING activationName = NULL,
					int activationConstraints = 0);

TCM_Task_Tree_Ref TCM_CreateMonitorNode (const TCM_Task_Tree_Ref &parentRef,
					 STRING name, 
					 const void *callbackData,
					 CALLBACK_FN_TYPE activationCallback,
					 unsigned int maxActivations = MAXINT,
					 unsigned int maxTriggers = MAXINT,
					 STRING activationName = NULL,
					 int activationConstraints = 0);

TCM_Task_Tree_Ref
TCM_CreatePollingMonitorNode (const TCM_Task_Tree_Ref &parentRef,
			      STRING name, 
			      const TCM_Action_Ref &activationAction,
			      MSecs period = seconds(1),
			      unsigned int maxActivations = MAXINT,
			      unsigned int maxTriggers = MAXINT,
			      STRING activationName = NULL,
			      int activationConstraints = 0,
			      BOOLEAN initialWait = FALSE);

TCM_Task_Tree_Ref
TCM_CreatePollingMonitorNode (const TCM_Task_Tree_Ref &parentRef,
			      STRING name, 
			      const void *callbackData,
			      CALLBACK_FN_TYPE activationCallback,
			      MSecs period = seconds(1),
			      unsigned int maxActivations = MAXINT,
			      unsigned int maxTriggers = MAXINT,
			      STRING activationName = NULL,
			      int activationConstraints = 0,
			      BOOLEAN initialWait = FALSE);

TCM_Return_Type TCM_Trigger (const TCM_Task_Tree_Ref &monitorRef);

TCM_Return_Type TCM_Activate (const TCM_Task_Tree_Ref &monitorRef,
			      const void *activationData);

TCM_Return_Type TCM_ActivateAt (const TCM_Task_Tree_Ref &monitorRef,
				TCM_Point const &activateTimePoint);
TCM_Return_Type TCM_ActivateAt (const TCM_Task_Tree_Ref &monitorRef,
				MSecs absoluteTime);
TCM_Return_Type TCM_ActivateIn (const TCM_Task_Tree_Ref &monitorRef,
				MSecs relativeWait);
// Activate this node in "relativeWait" msecs after 
//  the "activateTimePoint" have passed.
TCM_Return_Type TCM_ActivateInAfter (const TCM_Task_Tree_Ref &monitorRef,
				     MSecs relativeWait,
				     TCM_Point const &activateTimePoint);

const void *TCM_ActivationData (const TCM_Task_Tree_Ref &nodeRef);
unsigned int TCM_NumTriggers (const TCM_Task_Tree_Ref &nodeRef);
unsigned int TCM_NumActivations (const TCM_Task_Tree_Ref &nodeRef);

/*****************************************************************
 * 
 * Functions to invoke functions (not adding tree nodes) in 
 * response to particular events.
 *
 *****************************************************************/

TCM_Return_Type TCM_InvokeWhen (TCM_Point const &timePoint,
				SIMPLE_CALLBACK_FN_TYPE callback,
				const void *callbackData);

TCM_Return_Type TCM_InvokeWhen (MSecs absoluteTime,
				SIMPLE_CALLBACK_FN_TYPE callback,
				const void *callbackData);

TCM_Return_Type TCM_InvokeAfter (MSecs relativeWait,
				 SIMPLE_CALLBACK_FN_TYPE callback,
				 const void *callbackData);

/*****************************************************************
 * 
 * Types and functions for interacting with TCM
 *
 *****************************************************************/

TCM_Return_Type TCM_Initialize (void);

/****
 * If "waitingAllowed" is true, TCM_ProcessAgenda can block on
 * timeout signals and/or external events (otherwise, returns when
 * all pending internal events have been processed).
 ****/
TCM_Return_Type TCM_ProcessAgenda(BOOLEAN waitingAllowed  = TRUE,
				  MSecs   relativeTimeout = INFINITE_TIME );

typedef void (*SD_CALLBACK_FN_TYPE)(int sd, const void *callbackData);

TCM_Return_Type TCM_AddExternalEvent (int sd, SD_CALLBACK_FN_TYPE callback,
				      const void *callbackData);
	
TCM_Return_Type TCM_RemoveExternalEvent (int sd);

typedef void (* EXTERNAL_EVENT_BYPASS_HANDLER_TYPE)
				      ( BOOLEAN (*)(void *, struct timeval * ),
					void *,
					struct timeval *,
					BOOLEAN );
/* IE: 
 * void
 * ExternalEventBypassHandler ( BOOLEAN (* tcmDispatchExternalEventsHandler)
 *                                           ( void           * theTcmData,
 *					       struct timeval * theTimeout ),
 *				   void           * theTcmData,
 *				   struct timeval * theTimeout,
 *				   BOOLEAN          theHasTcmExternalEvents  );
 */
void TCM_SetExternalEventBypassHandler (
			 EXTERNAL_EVENT_BYPASS_HANDLER_TYPE theBypassHandler );

TCM_Return_Type TCM_DisplayTree(FILE *stream, const TCM_Task_Tree_Ref &nodeRef);

typedef enum { Log_None=0x0, Log_Msg=0x2, Log_Status=0x4, Log_Time=0x8,
	       Log_Id=0x10, Log_ParentId=0x20, 
	       Log_Ignore=0x40 } Logging_Options_Enum;

void TCM_SetTerminalLoggingOptions (int options);
void TCM_SetFileLoggingOptions (int options);

int TCM_TerminalLoggingOptions (void);
int TCM_FileLoggingOptions (void);

#ifdef DISTRIBUTED
TCM_Task_Tree_Ref TCM_AllocateDistributedNode (STRING agentName,
					       STRING taskName);

TCM_Return_Type TCM_SetDistributedAction (const TCM_Task_Tree_Ref &nodeRef, 
					  const void *callbackData);

// This is a shorthand for a series of allocate/setAction/insert calls
TCM_Task_Tree_Ref TCM_CreateDistributedNode(STRING agentName,
					    const TCM_Task_Tree_Ref &parentRef,
					    STRING taskName,
					    void *callbackData,
					    BOOLEAN isLastChild = TRUE);

typedef TCM_Task_Tree_Ref (*TASK_ALLOCATION_FN) (void);
typedef TCM_Action_Ref    (*ACTION_CREATION_FN) (const TCM_Task_Tree_Ref &node,
						 void *args);

TCM_Return_Type TCM_RegisterDistributedTask (STRING taskName, 
					     TASK_ALLOCATION_FN allocationFn,
					     ACTION_CREATION_FN actionFn,
					     STRING taskDataFormat);

/* Must be called once before connection distributed agents.
   "myHostMachine" defaults to the local machine. */
TCM_Return_Type TCM_EnableDistributedComm (STRING myName,
					   STRING myHostMachine = NULL);

/* Must be called for each agent that you want to send tasks to
   (or receive tasks from).
   "hostMachine" defaults to the local machine. */
TCM_Return_Type TCM_ConnectDistributedAgent (STRING agentName,
					     STRING hostMachine = NULL);

TCM_Return_Type TCM_RegisterDistributedException (STRING exceptionName, 
						  STRING exceptionDataFormat);

#endif

#endif /* INCtcm */
