/*****************************************************************************
 * 
 * PROJECT: Task Control Management.
 *
 * (c) Copyright 1996 Reid Simmons.  All rights reserved.
 *
 * FILE: tplConstr.h
 *
 * ABSTRACT: Classes for dealing with temporal constraints (essentially
 *           mapping between time points and signals/states).
 *
 * $Source: /afs/cs/project/TCA/Master/tcaV8/tcm/src/tplConstr.h,v $ 
 * $Revision: 1.5 $
 * $Date: 1999/08/04 14:00:19 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: tplConstr.h,v $
 * Revision 1.5  1999/08/04 14:00:19  reids
 * Distributed version of TCM.  Provides the ability for one process to
 *  add tasks, constraints, exception handlers, etc. with another process.
 *  The distributed version is created by compiling with -DDISTRIBUTED flag,
 *   which can be gotten by "gmaking" with DISTRIBUTED=1.
 *
 * Revision 1.4  1999/06/06 13:48:10  reids
 * Changes to TCM 1.3.0:
 *   New list data structures.
 *   New, generalized logging interface.
 *   Fixed firstChild to work if there are no children.
 *   Fixed code to deallocate a node.
 *   In certain circmumstances, don't send termination signal when task
 *     itself terminating.
 *
 * Revision 1.3  98/04/21  12:46:18  reids
 * Added InvokeWhen/InvokeAfter functionality, to invoke a function when a
 *   particular event occurs.
 * Added the DelayUntilAfter/TerminateAtAfter functionality, to delay/terminate
 *   a node by waiting some msecs after a particular event has occurred.
 * 
 * Revision 1.2  97/12/04  17:50:32  reids
 * Another fairly stable version (except that monitors do not quite work)
 * 
 * Revision 1.1  97/11/21  14:06:51  reids
 * First release of TCM -- seems to be a stable version
 * 
 *
 *****************************************************************************/

#ifndef INCtplConstr
#define INCtplConstr

#include "tcm.h"
#include "event.h"
#include "taskTree.h"

// Don't terminate unecessarily, that is:
// Don't add the timed event if the signal is start terminating,
//       and the node is in the terminating/terminated state,
//       and the signal state is not terminating/terminated.
#define ADD_EVENT_P(signal, afterNode, state) \
  !((signal) == Start_Terminating_Signal && \
    (afterNode)->validState(Terminating_State) && \
    !((state) == Terminating_State || (state) == Terminated_State))

class Time_Point
{
 public:
  Time_Point(TCM_Point_Enum whichPoint, TCM_Interval_Enum whichInterval,
	     Task_Tree_Ref const &whichNode)
    { point = whichPoint; interval = whichInterval; node = whichNode; }
  Time_Point(TCM_Point timePoint)
    { point = timePoint.point(); interval = timePoint.interval();
      node = timePoint.node(); }
  Task_Tree_Ref nodeOf(void) const { return node; }
  Signal_Enum signalOf(void) const;
  State_Enum stateOf(void) const;
  BOOLEAN isPast(void) const
    { return nodeOf()->validState(stateOf()); }

 private:
  TCM_Interval_Enum interval;
  TCM_Point_Enum point;
  Task_Tree_Ref node;
};

class After_Constraint
{ public:
  After_Constraint(Task_Tree_Ref theNode, Signal_Enum theSignal,
		   MSecs wait, Task_Tree_Ref theAfterNode, State_Enum theState)
    : node(theNode), afterNode(theAfterNode) 
    { waitTime = wait; signal = theSignal; state = theState; }

  Task_Tree_Ref node;
  Signal_Enum   signal;
  MSecs         waitTime;
  Task_Tree_Ref afterNode;
  State_Enum    state;
};
typedef After_Constraint *AFTER_CONSTRAINT_PTR;

void constrainBefore (const Time_Point &t1, const Time_Point &t2);

void constrainWhen (const Time_Point &tp, MSecs absoluteTime);

void constrainWhenAfter (Task_Tree_Ref signalledNode, Signal_Enum signal,
			 Time_Point afterPoint, MSecs relativeWait);

void terminateAfter (const Task_Tree_Ref &terminateRef, const Time_Point &tp);

void terminateWhen (const Task_Tree_Ref &terminateRef, MSecs absoluteTime);

BOOLEAN checkInconsistent(int tplConstraints);

#endif /* INCtplConstr */
