
import java.io.PrintStream;
import netscape.application.*;
import netscape.util.*;

/** GraphComponents form the abstract base class of all the Graph objects.
  * GraphComponents are, in the general case, lightweight Views that just know
  * how to draw themselves and/or their contained View.
  *
  * Note: There is nothing preventing different GraphComponents from
  * overwriting one-another.  If they do, we call it a "bug".
  *
  * GraphComponents are based around the idea that each Graph component can
  * have a maximum of 4 connections, one of which is the parent and the rest
  * of which are the children.  This default functionality can be overriden
  * in subclasses if it is desired. 
  *
  */
public abstract class GraphComponent extends Object
				  implements DragSource, GraphComponentSupplier
{
	/* Class variables */
  public static Color DEFAULT_FOREGROUND_COLOR = Color.black;
  public static Color DEFAULT_BACKGROUND_COLOR = Color.white;
  public static Color DEFAULT_SELECTION_COLOR  = Color.green;

    /* A Good place to establish constants. */
  public final static int NONE    = 0x00;
  public final static int LEFT    = 0x01;
  public final static int RIGHT   = 0x02;
  public final static int DOWN    = 0x04;
  public final static int UP      = 0x08;
  public final static int WEST    = LEFT;
  public final static int EAST    = RIGHT;
  public final static int SOUTH   = DOWN;
  public final static int NORTH   = UP;
  public final static int HORIZONTAL = LEFT | RIGHT;
  public final static int VERTICAL   = UP   | DOWN;
  public final static int POSITIVE   = UP   | LEFT;   /* parent source */
  public final static int NEGATIVE   = DOWN | RIGHT;  /* parent source */
  public final static int ALL        = LEFT | RIGHT | DOWN | UP;
  public final static int INVALID = -1;
  public final static int ILLEGAL = -1;


	/* Declare our drag-drop stuff... */
  public final static int     MOUSE_X_DRAG_SESSION_FUDGE_FACTOR = 3;
  protected    static Bitmap  dragDropImage = null;

	/* Class method to create/retrieve drag-drop image */
  public static Bitmap getDragDropImage()
  {
    if ( GraphComponent.dragDropImage == null )
    {
      PaletteView  paletteView = new PaletteView ( ".....", null, Color.red );
      GraphComponent.dragDropImage = paletteView . getImage();
    }
    return GraphComponent.dragDropImage;
  }

    /** Class method to reverse directions. */
  public static int invertDirection ( int theDirection )
  {
    switch ( theDirection )
    {
      case GraphComponent . LEFT:   return GraphComponent . RIGHT;
      case GraphComponent . RIGHT:  return GraphComponent . LEFT;
      case GraphComponent . DOWN:   return GraphComponent . UP;
      case GraphComponent . UP:     return GraphComponent . DOWN;
      default: return theDirection;
    }
  }
    /** Class method to test directions */
  public static boolean testDirection ( int theDirection, int theTestCriteria )
  {
    return ( ( theDirection & theTestCriteria ) != 0 );
  }
    /** Class method to add two directions together */
  public static int  addDirection ( int theDirection, int theOtherDirection )
  {
    return theDirection | theOtherDirection;
  }



    /** Coordinate of where this GraphComponent should paint itself.
      * Objects should cache their width & height here, to facilitate
      * the sizing computations of other objects.
      */
  protected Rect           componentBounds;

    /** Total rectangular area occupied by this object and all it's children.
      */
  protected Rect           subgraphBounds;

    /** Objects Color */
  protected Color          foregroundColor, backgroundColor, selectionColor;

    /** Object's parent object.  */
  protected GraphComponent parent;

    /** Object's offspring (children).  The number of children is
      * established when the constructor is called.
      * Note: Some of the children may be null.  There may be more
      * children in the array after a null child.
      */
  protected GraphComponent children[];

    /** The direction of the parent from this object's perspective. */
  protected int parentDirection;


    /** Is this child in the current selection? */
  protected boolean isSelected;

    /** Internal state variable used when multiple children realign
      * simultaneously.
      */
  private int disableRealignPropagationUpCount;

    /** GraphModule to which this graph-component belongs */
  protected GraphModule  owningGraphModule;


  GraphComponent ( int theNumberOfChildren )
  {
    int i;

    componentBounds = new Rect ( GraphComponent.INVALID, 
				 GraphComponent.INVALID,
				 0, 0 );
    subgraphBounds  = new Rect ( GraphComponent.INVALID, 
				 GraphComponent.INVALID,
				 0, 0 );
    setForegroundColor  ( null );
    setBackgroundColor  ( null );
    setSelectionColor   ( null );
    parent = null;
    children = new GraphComponent [ theNumberOfChildren ];
    for ( i=0;  i < children.length;  i++ )
      children[i] = null;
    parentDirection = GraphComponent.NONE;
    isSelected = false;
    disableRealignPropagationUpCount = 0;
    owningGraphModule = null;
  }



    /** This provides the default behavior for realign'ing.  It realigns this
      * object based on its parent's coordinates.  It assumes that this
      * object's width & height have already been set.  If necessary, this
      * method should be overriden in subclasses to provide width & height
      * information.  This method *MUST* invoke didRealign() after it
      * completes to trigger the realignment of the children.  
      */
  public void realign ( )
  {
	/* Assume that our width & height have already been set. */

    if (   ( getParent()               == null                   )
	|| ( getParent() . getX()      == GraphComponent.INVALID )
	|| ( getParent() . getY()      == GraphComponent.INVALID )
	|| ( getParent() . getWidth()  == GraphComponent.INVALID )
	|| ( getParent() . getHeight() == GraphComponent.INVALID ) )
    {
      setX ( GraphComponent.INVALID );
      setY ( GraphComponent.INVALID );
    }
    else
    {
      switch ( getParentDirection() )
      {
	case GraphComponent.RIGHT:
	  setX ( getParent() .  minX()  - 1 );
	  setY ( getParent() .  midY()  -  ( getHeight() / 2 ) );
	  break;

	case GraphComponent.LEFT:
	  setX ( getParent() .  maxX()  + 1 );
	  setY ( getParent() .  midY()  -  ( getHeight() / 2 ) );
	  break;

	case GraphComponent.DOWN:
	  setX ( getParent() .  midX()  -  ( getWidth() / 2 ) );
	  setY ( getParent() .  minY()  - 1);
	  break;

	case GraphComponent.UP:
	  setX ( getParent() .  midX()  -  ( getWidth() / 2 ) );
	  setY ( getParent() .  maxY()  +  1 );
	  break;

	default:
	  setX ( getParent() . maxX() + 1 );
	  setY ( getParent() . maxY() + 1 );
	  System.err.println ( "[GraphComponent:realign]  Warning:  "
			       + "Illegal value for getParentDirection().  ("
			       + getParentDirection() + ")" );
	  break;
      }
    }

	/* If we have realigned, our children also need to realign. */
    didRealign ();
  }



    /** This should be called after realign().  It realigns our children,
      * and triggers didrealignChildren().
      */
  public void didRealign ( )
  {
    int      i;

	/* Disable propagating realigns until we are done. */
    disableRealignPropagationUp();

	/* Realign our children. */
    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) != null )
      {
	getChild ( i ) . realign();
      }
    }

	/* Reenable propagating realigns now that our children are done... */
    reenableRealignPropagationUp();

        /* We need recompute & cache our subgraphBounds. */
	/* (And let our parent do the same.)             */
    didRealignChildren ( this, this );
  }




    /** Invoked when a descendent has changed.
      * Invokes parentDidRealignOtherChild() on other children with
      * theRealignedChildTree,
      * recomputes & caches this component's new subgraphBounds,
      * and notifies the parent to also update itself.
      *
      * Note: theRealignedChildTree is the branch (child) of this tree 
      *       that has changed.
      *
      * Note: theOriginalChild is used by GraphRoot/GraphView to specify
      *       the "Hot" area of the screen if we are in a ScrollView.
      */
  public void didRealignChildren ( GraphComponent theRealignedChildTree,
				   GraphComponent theOriginalChild      )
  {
    int i;
    
	/* Are we already doing a didRealignChildren?
	 * Ie: Did parentDidRealignOtherChild() or didRealign() trigger
	 *     a realignment of one of our children?  If so, scrub this.
	 */
    if ( getIsRealignPropagationUpDisabled() )
      return;

    	/* Disable propagating realigns until we are done. */
    disableRealignPropagationUp();


       /* We need recompute & cache our subgraphBounds. */    
    resetSubgraphBounds();
    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) != null )
      {
	   /* Let the other child update itself if necessary */
	getChild ( i ) . parentDidRealignOtherChild ( theRealignedChildTree );

	   /* And include its subgraphBounds in our subgraphBounds. */
	getSubgraphBounds()
	  . unionWith ( getChild ( i ) . getSubgraphBounds() );
      }
    }

	/* Perhaps our parent also wants to do this? */
    if ( getParent() != null )
      getParent() . didRealignChildren ( this, theOriginalChild );

	/* Reenable propagating realigns now that our children are done... */
    reenableRealignPropagationUp();
  }


    /** This method can be overriden in subclasses that want to know when
      * their parent's other children have been realigned.
      */
  public void parentDidRealignOtherChild (GraphComponent theRealignedChildTree)
  {
  }


    /** This can be invoked from any child to trigger a graph redraw */
  public void redisplay ( Rect theRectToRedisplay )
  {
    if ( getParent() != null )
      getParent() . redisplay ( theRectToRedisplay );
    else
      System.err.println ( "[GraphComponent:redisplay]  Error:  No Parent." );
  } 



    /** Overriden in children to actually Draw this object. 
      * Should not be invoked directly.  Instead, it is automatically
      * invoked by drawComponents. */ 
  public abstract void drawSelf ( Graphics theGraphics );



    /** The general way to draw this component and all its children. */
  public void drawComponents ( Graphics theGraphics )
  {
    int    i;

	/* Note: This would be a good place to establish a clipping-Rect. */

	/* Set our color */
    Color  originalColor = theGraphics . color();
    theGraphics . setColor ( getColor() );

	/* Draw ourself */
    drawSelf ( theGraphics );

	/* Draw our children */
    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) != null )
	getChild ( i ) . drawComponents ( theGraphics );
    }

	/* Restore the color back to what we originally had. */
    theGraphics . setColor ( originalColor );
  }


    /** Gets the Rect containing just this component's boundaries. */
  public Rect getComponentBounds ( )
  {
    if ( getWidth() < 0 )
    {
      setX     ( getX() + getWidth() );
      setWidth ( getWidth() * -1     );
    }

    if ( getHeight() < 0 )
    {
      setY      ( getY() + getHeight() );
      setHeight ( getHeight() * -1     );
    }

    return componentBounds;
  }

    /** Gets the Rect defining the area occupied by this object and
      * all it's children.
      */
  public Rect getSubgraphBounds ( )
  {
    if ( subgraphBounds . width < 0 )
    {
      subgraphBounds . x += subgraphBounds . width;
      subgraphBounds . width *= -1;
    }

    if ( subgraphBounds . height < 0 )
    {
      subgraphBounds . y += subgraphBounds . height;
      subgraphBounds . height *= -1;
    }

    return subgraphBounds;
  }

    /** Resets the subgraphBounds equal to the current compnent's boundaries */
  public void resetSubgraphBounds()
  {
    subgraphBounds . setBounds ( getComponentBounds() );
  }

    /** Convenience Method.  I was writing it in too many other places. */
  public boolean getChildrenSubgraph ( Rect            theBounds,
				       GraphComponent  excludingChild )
  {
    int      i;
    boolean  foundChild = false;

	/* Idiocy check */
    if ( theBounds == null )
    {
      System.err.println ( "[GraphComponent:getChildrenSubgraph]  Warning:  "
			   + "theBounds is null" );
      return false;
    }


    for ( i=0; i < getMaxChildren();  i++ )
    {
      if (   ( getChild ( i ) != null           )
	  && ( getChild ( i ) != excludingChild ) )
      {
	if ( foundChild == false )
	{
	  foundChild = true;
	  theBounds . setBounds ( getChild ( i ) . getSubgraphBounds() ); 
	}
	else
	  theBounds . unionWith ( getChild ( i ) . getSubgraphBounds() );
      }
    }

    return foundChild;
  }

	/** Compute the children subgraph the hard way, recursively,
	  * excluding any children with parentDirection's
	  * in excludingDirections */
  public boolean computeChildrenSubgraph ( Rect theBounds,
					   int  restrictToDirections )
  {
    return computeChildrenSubgraph ( theBounds, restrictToDirections, false );
  }

  protected boolean computeChildrenSubgraph ( Rect    theBounds,
					      int     restrictToDirections,
					      boolean hasSetBounds )
  {
    int      i;

	/* Idiocy check */
    if ( theBounds == null )
    {
      System.err.println ( "[GraphComponent:computeChildrenSubgraph]  "
			   + "Warning:  theBounds is null" );
      return false;
    }


    for ( i=0; i < getMaxChildren();  i++ )
    {
      if (   ( getChild ( i ) != null           )
	  && ( testDirection ( getChild ( i ) . getParentDirection(),
			       restrictToDirections ) ) )
      {
	if ( hasSetBounds == false )
	{
	  hasSetBounds = true;
	  theBounds . setBounds ( getChild ( i ) . getComponentBounds() );
	}
	else
	  theBounds . unionWith ( getChild ( i ) . getComponentBounds() );

	getChild ( i ) . computeChildrenSubgraph ( theBounds,
						   restrictToDirections,
						   true );
      }
    }

    return hasSetBounds;
  }


    /** Implemented by GraphComponents that contain Views */
  public void addSubviewsDown ( GraphRoot theRoot )
  {
    int i;
    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) != null )
	getChild ( i ) . addSubviewsDown ( theRoot );
    }
  }

  public void addSubviewsUp ( GraphComponent theComponent )
  {
    if ( getParent() != null )
      getParent() . addSubviewsUp ( theComponent );
  }

    /** Implemented by GraphComponents that contain Views */
  public void removeSubviewsDown ( GraphRoot theRoot )
  {
    int i;
    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) != null )
	getChild ( i ) . removeSubviewsDown ( theRoot );
    }
  }

  public void removeSubviewsUp ( GraphComponent theComponent )
  {
    if ( getParent() != null )
      getParent() . removeSubviewsUp ( theComponent );
  }



  public int   getX      (               ) { return componentBounds.x; }
  public void  setX      ( int theX      ) { componentBounds.x = theX; }

  public int   getY      (               ) { return componentBounds.y; }
  public void  setY      ( int theY      ) { componentBounds.y = theY; }

  public int   getWidth  (               ) { return componentBounds.width;    }
  public void  setWidth  ( int theWidth  ) { componentBounds.width = theWidth;}

  public int   getHeight (               ) { return componentBounds.height;   }
  public void  setHeight ( int theHeight ) { componentBounds.height=theHeight;}


  public int minX () { return getX();                       }
  public int midX () { return getX() + ( getWidth()  / 2 ); }
  public int maxX () { return getX() +   getWidth()  - 1;   }

  public int minY () { return getY();                       }
  public int midY () { return getY() + ( getHeight() / 2 ); }
  public int maxY () { return getY() +   getHeight() - 1;   }


  public int distanceToPointSquared ( int theX, int theY )
  {
    return distanceToPointSquared ( theX, theY, getComponentBounds() );
  }

	/** This really belongs in class Rect! */
  public static int distanceToPointSquared ( int theX, int theY, Rect theRect )
  {
    if ( theRect . contains ( theX, theY ) )
	 return 0;

    int dx = Math.min ( Math.max( theRect.x, theX ),
			theRect.x + theRect.width  - 1 ) - theX,
	dy = Math.min ( Math.max( theRect.y, theY ),
			theRect.y + theRect.height - 1 ) - theY;

    return (dx * dx) + (dy * dy);
  }


	/* Convenience methods to access the foreground color */
  public Color getColor  (               ) { return getForegroundColor();    }
  public void  setColor  ( Color theColor) { setForegroundColor( theColor ); }


  public Color getForegroundColor ( )
  {
    return (   getIsSelected() 
	     ? getSelectionColor()
	     : (   ( foregroundColor != null )
	         ? foregroundColor : GraphComponent.DEFAULT_FOREGROUND_COLOR )
	     );
  }
  public void  setForegroundColor ( Color theColor)
  {
    foregroundColor = theColor;
  }

  public Color getBackgroundColor ( )
  {
    return (   ( backgroundColor != null )
	     ? backgroundColor : GraphComponent.DEFAULT_BACKGROUND_COLOR );
  }
  public void  setBackgroundColor ( Color theColor)
  {
    backgroundColor = theColor;
  }

  public Color getSelectionColor ( )
  {
    return (   ( selectionColor != null )
	     ? selectionColor : GraphComponent.DEFAULT_SELECTION_COLOR );
  }
  public void  setSelectionColor ( Color theColor)
  {
    selectionColor = theColor;
  }


  public boolean getIsSelected ( ) { return isSelected;  }

  public void setSelected   ( boolean theIsOnlySelected,
			      boolean theShouldRedisplay   )
  {
    GraphComponent  component = this;

    if ( theIsOnlySelected )
    {
      while ( component . getParent() != null )
	component = component . getParent();
      component . setUnselected ( true, false );
    }

    isSelected = true;

    if ( theShouldRedisplay )
    {
      if ( theIsOnlySelected )
	redisplay ( null );
      else
	redisplay ( getComponentBounds() );
    }
  }

  public void setSelected   ( boolean theIsOnlySelected )
  {
    setSelected ( theIsOnlySelected, true );
  }

  public void setSelected   ( )
  {
    setSelected ( false, true );
  }


  public void setUnselected ( boolean theIncludingChildren,
			      boolean theShouldRedisplay   )
  {
    isSelected = false;

    if ( theIncludingChildren )
    {
      for ( int i=0;  i < getMaxChildren();  i++ )
      {
	if ( getChild ( i ) != null )
	  getChild ( i ) . setUnselected ( true, false );
      }
    }

    if ( theShouldRedisplay )
    {
      if ( theIncludingChildren )
	redisplay ( getSubgraphBounds()  );
      else
	redisplay ( getComponentBounds() );
    }
  }


  public void setUnselected ( )
  {
	/* Only redisplay if we are changing from selected to unselected. */
    setUnselected ( false, getIsSelected() );
  }


  public void findSelectedComponents ( Vector  theVector )
  {
    if ( getIsSelected() )
      theVector . addElement ( this );

    for ( int i=0;  i < getMaxChildren();  i++ )
    {
      if ( getChild ( i ) != null )
	getChild ( i )  . findSelectedComponents ( theVector );
    }
  }



  public boolean getIsRealignPropagationUpDisabled ( ) 
			       { return disableRealignPropagationUpCount > 0; }
  public void    disableRealignPropagationUp  ( )
			       { disableRealignPropagationUpCount++; }
	/** Note: Any didRealignChildren() events that occured between
	  * the first (begin-nesting) disableRealignPropagationUp() invocation
	  * and the last (end-nesting) reenableRealignPropagationUp()
	  * have been lost.  Therefore: You should follow
	  * reenableRealignUpPropagationUp() with the appropriate
	  * didRealignChildren() to properly update the graph.
	  */
  public void    reenableRealignPropagationUp ( )
  {
    if ( disableRealignPropagationUpCount > 0 )
      disableRealignPropagationUpCount--;
  }



  public void setOwningGraphModule ( GraphModule  theGraphModule )
  {
    owningGraphModule = theGraphModule;
  }

  private GraphModule getActualValueOfOwningGraphModule()
						  { return owningGraphModule; }

	/* Returns the GraphModule immediately above us */
	/* (Presumably the one we are a part of...) */ 
  public GraphModule getOwningGraphModule()
  {
    if ( owningGraphModule == null )
    {
      if ( getParent() != null )
      {
	return getParent() . getOwningGraphModule();
      }
      else
      {
	System.err.println( "[GraphComponent:getOwningGraphModule]  Warning:  "
			    + "Both owningGraphModule & parent are NULL!" );
      }
    }
    return owningGraphModule;
  }



  public GraphRoot      getGraphRoot ( )
  {
    GraphComponent   ancestor = getParent();

    while (   (  ancestor                       != null  )
	   && ( (ancestor instanceof GraphRoot) == false ) )
    {
      ancestor = ancestor . getParent();
    }

    if ( ancestor == null )
      return null;
    else
      return (GraphRoot) ancestor;
  }


  public GraphComponent getParent  ( )     { return parent; }

  public    void        setParent ( GraphComponent theParent,
				    boolean        theShouldRemove,
				    boolean        theShouldAdd,
				    boolean        theShouldRealign )
  {
    if ( theParent != getParent() )
    {
      if ( theShouldRemove && ( getParent() != null ) )
	getParent() . removeChild ( this );

      parent = theParent;

      if ( theShouldAdd )
	theParent . addChild ( this );

      if ( theShouldRealign )
	realign();
    }
  }

  public    void        setParent  ( GraphComponent theParent  )
  {
    setParent ( theParent, true, true, true );
  }

  public boolean getIsAncestor ( GraphComponent  theGraphComponent )
  {
    GraphComponent   currentGraphComponent = this;

    while (   (  currentGraphComponent                       != null  )
	   && ( (currentGraphComponent instanceof GraphRoot) == false ) )
    {
      if ( currentGraphComponent == theGraphComponent )
	return true;

      currentGraphComponent = currentGraphComponent . getParent();
    }

    return false;
  }



  public int getMaxChildren() { return children.length; }

  public int getCountChildren ( )
  {
    int i, count=0;

    for ( i=0; i < getMaxChildren(); i++ )
    {
       if ( getChild ( i ) != null )
	 count ++;
    }

    return count;
  }

  public GraphComponent getChild ( int theIndex )
  {
    if ( ( theIndex >= 0 )  &&  ( theIndex < getMaxChildren() ) )
      return children [ theIndex ];
    else
    {
      System.err.println ( "[GraphComponent:getChild]  Error:  "
			   + "theIndex (" + theIndex + ") is out of range." );
      return null;
    }
  }

  public GraphComponent setChild ( int theIndex, GraphComponent theChild )
  {

    if ( ( theIndex >= 0 )  &&  ( theIndex < getMaxChildren() ) )
    {
      GraphComponent  oldChild = children [ theIndex ];
      children [ theIndex ] = theChild;
      return oldChild;
    }
    else
    {
      System.err.println ( "[GraphComponent:setChild]  Error:  "
			   + "theIndex (" + theIndex + ") is out of range." );
      return null;
    }
  }

//  public GraphComponent[]  getChildren ( ) { return children; }
  public void setChildren ( GraphComponent theChildren[] )
  {
    children = theChildren;
    realign();
  }

    /** The direction of the parent from this object's perspective. */
  public int  getParentDirection ( )  { return parentDirection; }
  public void setParentDirection ( int theParentDirection )
  {
    if ( parentDirection != theParentDirection )
    {
      GraphComponent parent = getParent();

      if (   ( parent                        != null  )
	  && ( parent . removeChild ( this ) == false ) )
      {
	System.err.println ( "[GraphComponent:setParentDirection]  Warning: "
			     + "Was not a child of parent." );
      }

      parentDirection = theParentDirection;

      if (   ( parent                     != null  )
	  && ( parent . addChild ( this ) == false ) )
      {
	System.err.println ( "[GraphComponent:setParentDirection]  Error: "
			     + "Unable to add child back to parent.  "
			     + "Child left disconnected." );
      }

      

    }
  }



  public int findChildIndex ( GraphComponent theChild )
  {
    int i;

    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) == theChild )
      {
	return i;
      }
    }
    return GraphComponent.INVALID;
  }

	/** Find the first child with parentDirection in theParentDirection */
  public int findChildIndex ( int theParentDirection )
  {
    int i;

    for ( i=0; i < getMaxChildren(); i++ )
    {
      if (   ( getChild ( i ) != null )
	  && ( testDirection ( getChild ( i ) . getParentDirection(),
			       theParentDirection ) ) )
      {
	return i;
      }
    }
    return GraphComponent.INVALID;
  }

  public GraphComponent findChild ( int theParentDirection )
  {
    int index = findChildIndex ( theParentDirection );

    if ( index != GraphComponent.INVALID )
      return getChild ( index );
    else
      return null;
  }



  public boolean addChild ( GraphComponent theComponent )
  {
    int i;

    if (     invertDirection ( getParentDirection() )
	  ==    theComponent . getParentDirection()   )
    {
      System.err.println ( "[GraphComponent:addChild]  Error: "
			   + "Direction already taken by parent.   ("
			   + theComponent . getParentDirection() + ")" );
      return false;
    }

    if (    findChildIndex ( theComponent . getParentDirection() )
	 != GraphComponent.INVALID )
    {
      System.err.println ( "[GraphComponent:addChild]  Error: "
			   + "Direction already taken.   ("
			   + theComponent . getParentDirection() + ")" );
      return false;
    }


    for ( i=0; i < getMaxChildren(); i++ )
    {
      if ( getChild ( i ) == null )
      {
	setChild ( i, theComponent );
	addSubviewsUp ( theComponent );
	theComponent . setParent ( this, true, false, true );
	didAddChild ( theComponent );
	return true;
      }
    }

    System.err.println ( "[GraphComponent:addChild]  Error: "
			 + "All Children slots filled." );
    return false;
  }


  public boolean removeChild ( int i )
  {
    if (   ( i >= 0 )
	&& ( i < getMaxChildren() )
	&& ( getChild ( i ) != null ) )
    {
      GraphComponent removedChild = getChild ( i );
      setChild ( i, null );
      removedChild . setParent ( null, false, false, false );
      removeSubviewsUp ( removedChild );
      didRemoveChild ( removedChild );
	/* Note that, while we have not realigned, and our still-existing
	 * children have not realigned, our sub-tree has effectively realigned.
	 */ 
      didRealignChildren ( this, this );
      return true;
    }
    else
    {
      return false;
    }
  }

  public boolean removeChild ( GraphComponent theComponent )
  {
    return removeChild ( findChildIndex ( theComponent ) );
  }


  public void didAddChild ( GraphComponent theComponent )
  {
  }

  public void didRemoveChild ( GraphComponent theComponent )
  {
  }



  public GraphComponent findComponent ( int theX, int theY )
  {
	/* If it is inside me. */
    if ( getComponentBounds() . contains ( theX, theY ) )
      return this;

	/* if it is not inside my children. */
    if ( getSubgraphBounds() . contains ( theX, theY ) == false )
      return null;

    for ( int i=0; i < getMaxChildren(); i++ )
    {
       if (   ( getChild ( i ) != null )
	   && ( getChild ( i ) . getSubgraphBounds()
			       . contains ( theX, theY ) ) )
       {
	 return getChild ( i ) . findComponent ( theX, theY );
       }
    }

    return null;
  }


  public GraphComponent findNearestComponent ( int theX, int theY )
	{ return findNearestComponent ( theX, theY, GraphComponent.INVALID ); }

  public GraphComponent findNearestComponent ( int theX, int theY,
					       int theClosestDistanceSoFar )
  {
    int            childDistance;
    GraphComponent closestComponent = null,
                   childComponent   = null;

	/* If it is inside me. */
    if ( getComponentBounds() . contains ( theX, theY ) )
      return this;  /* Distance=0 */

	/* How close are we to this point? */
    if (   ( theClosestDistanceSoFar == GraphComponent.INVALID )
	|| ( distanceToPointSquared( theX, theY ) < theClosestDistanceSoFar ) )
    {
      theClosestDistanceSoFar = distanceToPointSquared ( theX, theY );
      closestComponent = this;
    }

	/* if our subgraph is no closer */
    if ( distanceToPointSquared ( theX, theY, getSubgraphBounds() )
	 >= theClosestDistanceSoFar )
    {
      return closestComponent;
    }

    for ( int i=0; i < getMaxChildren(); i++ )
    {
       if (   ( getChild ( i ) != null )
	   && (   theClosestDistanceSoFar
		> distanceToPointSquared ( theX, theY,
					   getChild ( i )
					     . getSubgraphBounds() ) ) )
       {
	 childComponent = getChild ( i ) . findNearestComponent ( theX, theY );
	 
	 if ( childComponent != null )
	 {
	   childDistance = childComponent . distanceToPointSquared ( theX,
								     theY );
	   if ( childDistance < theClosestDistanceSoFar )
	   {
	     closestComponent        = childComponent;
	     theClosestDistanceSoFar = childDistance;
	   }
	 }
       }
    }

    return closestComponent;
  }




  public boolean destroySubtree()
  {
    boolean  returnValue = true;

    if (   ( getActualValueOfOwningGraphModule() != null )
        && ( getActualValueOfOwningGraphModule() . unregisterToObserve()
	     == false ) )
      returnValue = false;

    for ( int i=0;  i < getMaxChildren();  i++ )
    {
      if ( getChild ( i ) != null )
      {
	if ( getChild ( i ) . destroySubtree() == false )
	  returnValue = false;
      }
    }

    return returnValue;
  }



  public int getGraphType()
  {
    if ( getParent() != null )
      return getParent() . getGraphType();
    else
    {
      System.err.println ( "[GraphComponent:getGraphType] Error: "
			   + "No Parent to get Graph Type from..." );
      return GraphRoot.ILLEGAL_GRAPH;
    }
  }




  public boolean isLeftMouseButtonDown   ( MouseEvent theEvent )
  {
    return (   ( theEvent . isAltKeyDown()  == false )
	    && ( theEvent . isMetaKeyDown() == false ) );
  }
  public boolean isMiddleMouseButtonDown ( MouseEvent theEvent )
			  { return theEvent . isAltKeyDown(); }

  public boolean isRightMouseButtonDown  ( MouseEvent theEvent )
			  { return theEvent . isMetaKeyDown(); }

  public GraphComponent mouseDown ( MouseEvent theEvent )
  {
    if (   ( isLeftMouseButtonDown   ( theEvent ) )
	|| ( isMiddleMouseButtonDown ( theEvent ) ) )
    {
      GraphModule  ourGraphModule = getOwningGraphModule();

      if ( ourGraphModule != null )
      {
	ourGraphModule . toggleSelected ( theEvent . isShiftKeyDown(), this );
      }
      else
      {
	if ( getIsSelected() && theEvent . isShiftKeyDown() )
	{
	  setUnselected();
	}
	else
	{
	  setSelected ( theEvent . isShiftKeyDown() == false );
	}
      }
    }

    if (   ( isMiddleMouseButtonDown ( theEvent )   )
	&& ( getOwningGraphModule() != null           )
	&& ( getOwningGraphModule() . getIsSelected() ) )
    {
      GraphRoot  graphRoot = getGraphRoot();
      int        xAdjust, yAdjust;

      if ( graphRoot != null )
      {
	xAdjust = graphRoot . getSubviewXAdjust();
	yAdjust = graphRoot . getSubviewYAdjust();
      }
      else
      {
	xAdjust = 0;
	yAdjust = 0;
      }

      new DragSession ( this, GraphComponent.getDragDropImage(),

			theEvent.x
			  + GraphComponent.MOUSE_X_DRAG_SESSION_FUDGE_FACTOR
			  + xAdjust,

			theEvent.y
			  - ( GraphComponent.getDragDropImage().height() / 2 )
			  + yAdjust,

			theEvent.x + xAdjust,

			theEvent.y + yAdjust,

			"GraphComponentSource",

			this );

      return null; /* Must return null if doing a DragSession() */
    }

    return null;
  }

  public void mouseDragged ( MouseEvent theEvent )
  {
  }

  public void mouseUp ( MouseEvent theEvent )
  {
  }

  public void subviewDidResize ( View theSubview )
  {
  }


	/* DragSource Interface */
  public void dragWasAccepted ( DragSession theDragSession )
  {
    GraphModule  ourGraphModule = getOwningGraphModule();

    if ( ourGraphModule != null )
      ourGraphModule . removeSelf ( false, true );
    else
    {
      System.err.println ( "[GraphComponent.dragWasAccepted]  Warning:  "
			   + "ourGraphModule == null." );
    }
  }

  public boolean dragWasRejected ( DragSession theDragSession )
  {
    return true;
  }

  public View sourceView ( DragSession theDragSession )
  {
    if ( getParent() != null )
      return getParent() . sourceView ( theDragSession );
    else
    {
      System.err.println ( "[GraphComponent.sourceView]  Warning:  "
			   + "getParent() == null." );
      return null;
    }
  }
	/* Personal Drag & Drop extension interface */
  public DragDestination acceptsDrag ( DragSession theDragSession,
				       int theX, int theY )
  {
    return null;
  }

	/* GraphComponentSupplier interface */
  public boolean       supportsGraphType ( int theGraphType )
  {
    return ( theGraphType == getGraphType() );
  }

  public GraphModule   createGraphModule ( int theGraphType )
  {
    GraphModule   ourGraphModule = getOwningGraphModule();

    if ( ourGraphModule == null )
    {
      System.err.println ( "[GraphComponent.createGraphModule]  Warning:  "
			   + "ourGraphModule == null." );
      return null;
    }

    if ( ourGraphModule . getDataComponent() == null )
    {
      System.err.println ( "[GraphComponent.createGraphModule]  Warning:  "
			   + "ourGraphModule . getDataComponent() == null." );
      return null;
    }

    return ourGraphModule . getDataComponent()
                          . createGraphModule ( theGraphType );
  }

	/* Note:  returns a COPY of our DataComponent */
  public DataComponent getDataComponent ( )
  {
    GraphModule   ourGraphModule = getOwningGraphModule();
    DataComponent copyDataComponent;

    if ( ourGraphModule == null )
    {
      System.err.println ( "[GraphComponent.getDataComponent]  Warning:  "
			   + "ourGraphModule == null." );
      return null;
    }

    if ( ourGraphModule . getDataComponent() == null )
    {
      System.err.println ( "[GraphComponent.getDataComponent]  Warning:  "
			   + "ourGraphModule . getDataComponent() == null." );
      return null;
    }

    copyDataComponent = ourGraphModule . getDataComponent() . copyComponent();

    if ( copyDataComponent == null )
    {
      System.err.println ( "[GraphComponent.getDataComponent]  Warning:  "
			   + "copyDataComponent == null." );
    }

    return copyDataComponent;
  }





    /** For Debugging/Testing..  Its nice to see what things are... */
  public void print ( PrintStream theStream, boolean doRecursive )
  {
    theStream . println ( "componentBounds...: " + getComponentBounds() );
    theStream . println ( "subgraphBounds....: " + getSubgraphBounds()  );
    theStream . println ( "color.............: " + getColor() );
    theStream . println ( "disableRealignPUpC: "
			  + disableRealignPropagationUpCount );
    theStream . println ( "id................: 0x"
			  + Integer.toHexString ( hashCode() )
			  + "   [ " + getClass() . getName() + " ]" );
    theStream . println ( "parent............: 0x"
      + (   ( getParent() != null )
	  ? ( "" + Integer.toHexString ( getParent() . hashCode() )
	      + "   [ " + getParent() . getClass() . getName() + " ]" )
	  : "null" ) );
    theStream . println ( "ParentDirection...: "
	  + ( (getParentDirection() == GraphComponent.RIGHT ) ? "RIGHT" :
	      (getParentDirection() == GraphComponent.LEFT  ) ? "LEFT"  :
	      (getParentDirection() == GraphComponent.DOWN  ) ? "DOWN"  :
	      (getParentDirection() == GraphComponent.UP    ) ? "UP"    :
	      (getParentDirection() == GraphComponent.NONE  ) ? "***NONE***" :
	      "UNKNOWN" ) );
    int i;
    for ( i = 0;  i < getMaxChildren();  i++ )
      theStream . println ( "  child [" + i + "].....: " 
	+ (   ( getChild ( i ) != null )
	    ? ( "" + Integer.toHexString ( getChild ( i ) . hashCode() )
		+ "   [ " + getChild ( i ) . getClass() . getName() + " ]" )
	    : "null" ) );
    theStream . println ( "" );
    for ( i = 0;  doRecursive && ( i < getMaxChildren() );  i++ )
      if ( getChild ( i ) != null )
      {
	theStream . println ( "  child [" + i + "] ... of "
			      + Integer.toHexString ( hashCode() ) );
	getChild ( i ) . print ( theStream, doRecursive );
      }
  }

  public void print ( PrintStream theStream ) { print ( theStream,  false ); }
  public void print (                       ) { print ( System.out, false ); }
}
